package net.pterodactylus.irc;

import java.io.IOException;

/**
 * A connection to an IRC server.
 *
 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
 */
public interface Connection {

	/**
	 * Returns the hostname of the remote end of the connection.
	 *
	 * @return The remote’s hostname
	 */
	String hostname();

	/**
	 * Returns the port number of the remote end of the connection.
	 *
	 * @return The remote’s port number
	 */
	int port();

	/**
	 * Returns whether this connection has already been established.
	 *
	 * @return {@code true} as long as this connection is established, {@code
	 * false} otherwise
	 */
	boolean established();

	/**
	 * Returns the nickname that is currently in use by this connection. The
	 * nickname is only available once the connection has been {@link #start()}ed.
	 *
	 * @return The current nickname
	 */
	String nickname();

	/**
	 * Sets the username to use.
	 *
	 * @param username
	 * 		The username to use
	 * @return This connection
	 */
	Connection username(String username);

	/**
	 * Sets the real name to use.
	 *
	 * @param realName
	 * 		The real name to use
	 * @return This connection
	 */
	Connection realName(String realName);

	/**
	 * Sets the optional password for the connection.
	 *
	 * @param password
	 * 		The password for the connection
	 * @return This connection
	 */
	Connection password(String password);

	/**
	 * Returns the current rate of the connection’s incoming side.
	 *
	 * @return The current input rate (in bytes per second)
	 */
	long getInputRate();

	/**
	 * Returns the current rate of the connection’s outgoing side.
	 *
	 * @return The current output rate (in bytes per second)
	 */
	long getOutputRate();

	/**
	 * Checks whether the given source is the client represented by this
	 * connection.
	 *
	 * @param source
	 * 		The source to check
	 * @return {@code true} if this connection represents the given source, {@code
	 * false} otherwise
	 */
	boolean isSource(Source source);

	/**
	 * Joins the given channel.
	 *
	 * @param channel
	 * 		The channel to join
	 * @throws IOException
	 * 		if an I/O error occurs
	 */
	void joinChannel(String channel) throws IOException;

	/**
	 * Sends a message to the given recipient, which may be a channel or another
	 * nickname.
	 *
	 * @param recipient
	 * 		The recipient of the message
	 * @param message
	 * 		The message
	 * @throws IOException
	 * 		if an I/O error occurs
	 */
	void sendMessage(String recipient, String message) throws IOException;

	/**
	 * Sends a DCC RESUME request to the given recipient.
	 *
	 * @param recipient
	 * 		The recipient of the request
	 * @param filename
	 * 		The name of the file to resume
	 * @param port
	 * 		The port number from the original DCC SEND request
	 * @param position
	 * 		The position at which to resume the transfer
	 * @throws IOException
	 * 		if an I/O error occurs
	 */
	void sendDccResume(String recipient, String filename, int port, long position)
			throws IOException;

	/**
	 * Opens the connection and starts processing messages from the other side.
	 */
	void open();

	/**
	 * Closes this connection.
	 *
	 * @throws IOException
	 * 		if an I/O error occurs
	 */
	void close() throws IOException;

}
