/*
 * XdccDownloader - Bot.java - Copyright © 2013 David Roden
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.pterodactylus.xdcc.data;

import static com.google.common.base.Preconditions.checkNotNull;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import com.google.common.collect.Maps;

/**
 * A bot is a client in a {@link Network} that carries {@link Pack}s which are
 * available for download.
 *
 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
 */
public class Bot implements Iterable<Pack> {

	/** The network this bot is on. */
	private final Network network;
	private final String channel;

	/** The packs this bot carries. */
	private final Map<String, Pack> packs = Maps.newHashMap();

	/** The current name of the bot. */
	private String name;

	public Bot(Network network, String channel, String name) {
		this.network = checkNotNull(network, "network must not be null");
		this.channel = checkNotNull(channel, "channel must not be null");
		this.name = checkNotNull(name, "name must not be null");
	}

	//
	// ACCESSORS
	//

	/**
	 * Returns the network the bot is on.
	 *
	 * @return The network the bot is on
	 */
	public Network network() {
		return network;
	}

	public String channel() {
		return channel;
	}

	/**
	 * Returns the current name of this bot.
	 *
	 * @return The current name of this bot
	 */
	public String name() {
		return name;
	}

	/**
	 * Returns the packs this bot carries.
	 *
	 * @return The packs this bot carries
	 */
	public Collection<Pack> packs() {
		return packs.values();
	}

	//
	// MUTATORS
	//

	/**
	 * Sets the current name of this bot.
	 *
	 * @param name
	 * 		The name of this bot
	 * @return This bot
	 */
	public Bot name(String name) {
		this.name = checkNotNull(name, "name must not be null");
		return this;
	}

	//
	// ACTIONS
	//

	/**
	 * Adds the given pack to this bot.
	 *
	 * @param pack
	 * 		The pack to add
	 */
	public void addPack(Pack pack) {
		packs.put(pack.id(), pack);
	}

	//
	// ITERABLE METHODS
	//

	@Override
	public Iterator<Pack> iterator() {
		return packs.values().iterator();
	}

	//
	// OBJECT METHODS
	//

	@Override
	public boolean equals(Object object) {
		if (!(object instanceof Bot)) {
			return false;
		}
		Bot bot = (Bot) object;
		return network().equals(bot.network()) && name().equals(bot.name());
	}

	@Override
	public int hashCode() {
		return network().hashCode() ^ name().hashCode();
	}

	@Override
	public String toString() {
		return String.format("%s/%s", name(), network().name());
	}

}
