/*
 * XdccDownloader - StatsCommand.java - Copyright © 2013 David Roden
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.pterodactylus.xdcc.ui.stdin;

import static java.lang.String.format;
import static java.util.Collections.emptyList;

import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.List;
import java.util.StringJoiner;

import net.pterodactylus.xdcc.core.Core;
import net.pterodactylus.xdcc.data.ConnectedNetwork;

/**
 * Command that outputs a short statistic of what is going on.
 *
 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
 */
public class StatsCommand implements Command {

	private final Core core;

	public StatsCommand(Core core) {
		this.core = core;
	}

	@Override
	public String getName() {
		return "stats";
	}

	@Override
	public Collection<String> getAliases() {
		return emptyList();
	}

	@Override
	public State execute(State state, List<String> parameters, Writer outputWriter) throws IOException {
		int totalBotCount = 0;
		int totalPackCount = 0;
		for (ConnectedNetwork network : core.connectedNetworks()) {
			dumpNetworkStats(outputWriter, network);
			totalBotCount += network.getBotCount();
			totalPackCount += network.getPackCount();
		}
		outputWriter.write(format("Total: %d bots, %d packs.\n", totalBotCount, totalPackCount));
		return state;
	}

	private void dumpNetworkStats(Writer outputWriter,
			ConnectedNetwork network) throws IOException {
		outputWriter.write(format("Connected to %s via %s:%d.\n",
				network.getNetwork().name(), network.getHostname(),
				network.getPort(), network.getNickname()));
		String joinedChannels = joinChannelNames(network.getChannels());
		String forcedChannels = joinChannelNames(network.getForcedChannels());
		if (forcedChannels.isEmpty()) {
			if (!joinedChannels.isEmpty()) {
				outputWriter.write(format("  Joined %s.\n", joinedChannels));
			}
		} else {
			if (joinedChannels.isEmpty()) {
				outputWriter.write(
						format("  Force-joined %s.\n", forcedChannels));
			} else {
				outputWriter.write(format("  Joined %s, force-joined %s.\n",
						joinedChannels, forcedChannels));
			}
		}
		outputWriter.write(
				format("  %d bots serving %d packs.\n", network.getBotCount(),
						network.getPackCount()));
	}

	private String joinChannelNames(Collection<String> channels) {
		StringJoiner joinedChannels = new StringJoiner(", ");
		channels.stream().forEach((channel) -> joinedChannels.add(channel));
		return joinedChannels.toString();
	}

}
